from symbolic import equivalence_checking, compute_functional_summary, verify_output_true
from z3 import *


def my_max(a: int, b: int):
    if a >= b:
        return a
    else:
        return b


def fibonacci1(n: int):
    if n <= 0 or n > 10:
        return 0

    num1 = 0
    num2 = 1
    next_number = num2
    count = 1

    while count <= n:
        count += 1
        num1, num2 = num2, next_number
        next_number = num1 + num2

    return num1


def fibonacci2(n: int):
    if n <= 0 or n > 10:
        return 0
    elif n == 1:
        return 1
    else:
        return fibonacci2(n - 1) + fibonacci2(n - 2)


def get_sum(lst: list):
    result = 0
    for i in lst:
        result = i + result

    return result


def get_sum_alt(lst: list):
    return sum(lst)


def bounded_gcd_alt(x: int, y: int):
    if x > 10 or y > 10 or x <= 0 or y <= 0:
        return 0
    else:
        return gcd_alt(x, y)


def gcd_alt(x, y):
    if x == y:
        return x
    elif y < x:
        return gcd_alt(x - y, y)
    else:
        return gcd_alt(x, y - x)


def bounded_gcd(x: int, y: int):
    if x > 10 or y > 10 or x <= 0 or y <= 0:
        return 0
    else:
        return gcd(x, y)


def gcd(x, y):
    if x == y:
        return x
    elif y > x:
        return gcd(x, y - x)
    else:
        return gcd(x - y, y)


def my_sort(lst: list):
    sorted_result = sorted(lst)
    if lst.__len__() > 0:
        return min(lst) == sorted_result[0]
    else:
        return True


def double_sort(lst: list):
    l1 = sorted(lst)
    l2 = sorted(l1)
    return l1 == l2


def no_less_than(lst: list, target):
    result = 0
    for l in lst:
        if l >= target:
            result += 1

    return result


def less_than(lst: list, target: int):
    result = 0
    for l in lst:
        if l < target:
            result += 1
    return result


def check_comparison(lst: list, target: int):
    return lst.__len__() == no_less_than(lst, target) + less_than(lst, target)


def test_eq():
    assert equivalence_checking(bounded_gcd, bounded_gcd_alt, int)
    assert not equivalence_checking(bounded_gcd, my_max, int)
    assert equivalence_checking(fibonacci1, fibonacci2, int)


def test_summary():
    # first compute the summary for my_max
    summary, f_in, f_out = compute_functional_summary(my_max, int, "output")
    s = Solver()
    s.append(summary)
    # we can verify that the output of my_max is always an upper-bound of the inputs
    s.append(Not(And([f_out >= i for i in f_in])))
    assert s.check() == unsat


def test_list():
    verify_output_true(my_sort)
    verify_output_true(double_sort)
    verify_output_true(check_comparison)


if __name__ == "__main__":
    summary, f_in, f_out = compute_functional_summary(my_max, int, "output", print_summary=True)
    print(summary)

    summary, f_in, f_out = compute_functional_summary(fibonacci1, int, "output")
    print(summary)

    test_summary()

    # uncomment after finishing task 1
    # test_eq()

    # uncomment after finishing task 2
    # test_list()
